const uncapitalize = require('../utils/string/uncapitalize');
const capitalize = require('../utils/string/capitalize');
const lowercase = require('../utils/string/lowercase');
const pluralize = require('pluralize');
const { singular } = pluralize;
const quote = require('../utils/string/quote');
const removeUndefined = require('../utils/object/removeUndefined');
const primitiveGraphQLTypes = require('./primitiveGraphQLTypes');
const orms = require('../orms');
// Find a nesting context for current argument
const root = (fields, nestingContext) => {
  nestingContext.forEach((name) => {
    fields = fields.find((f) => (f.name === name) && f.isObject).fields;
  });
  return fields;
};

const primitiveTypes = (ormSetting) =>
  ormSetting.supportedPrimitiveTypes.concat(['File', 'Enum']);

module.exports = (args, ormSetting) => {

  // Set default orm setting
  if (!ormSetting) {
    ormSetting = orms['mongoose'];
  }

  // Model name and collection name
  const arg0 = args.shift();
  if (!arg0.match(/^[a-zA-Z]\w*(\/[a-zA-Z]\w*)?$/)) {
    throw `Unexpected Model/pluralVariableName '${arg0}'.`;
  }
  let [modelName, pluralVarName] = arg0.split('/');
  const varName = uncapitalize(modelName);
  !pluralVarName && (pluralVarName = pluralize(varName));
  modelName = capitalize(modelName);
  const collectionName = lowercase(pluralVarName);
  if (pluralVarName === varName) {
    pluralVarName = `all${capitalize(pluralVarName)}`;
  }

  // Fields
  const fields = [];
  const nestingContext = [], singularNestingContext = []; // for nested

  args.forEach((arg) => {
    const tokens = arg.split(':');
    if (tokens.length > 3) {
      throw `Unexpected field descriptor '${arg}'.`;
    }

    // Nesting Structure
    if (tokens[1] && tokens[1].match(/^\[?{$/)) {
      const isArray = !!tokens[1].match(/^\[{$/);
      root(fields, nestingContext).push({
        name: tokens[0],
        isObject: true,
        isArray,
        fields: []
      });
      nestingContext.push(tokens[0]);
      singularNestingContext.push(isArray ? singular(tokens[0]) : tokens[0]);
      return;
    } else if (tokens[0].match(/}\]?/)) {
      nestingContext.pop();
      singularNestingContext.pop();
      return;
    }

    // Schema referencing
    if (tokens[1] && tokens[1].match(/^\[?.*Schema\]?$/)) {
      const isArray = !!tokens[1].match(/\[.*\]/);
      const type = capitalize(tokens[1].match(/^\[?(.*)Schema\]?$/)[1]);
      const jsType = uncapitalize(type) + 'Schema';
      root(fields, nestingContext).push({
        name: tokens[0],
        jsType,
        graphQLType: type,
        type: type,
        isArray,
        isSchema: true
      });
      return;
    }

    // Default to String type if no type is specified
    tokens[1] || (tokens[1] = 'String');

    const name = uncapitalize(tokens[0]);
    let modifier = '', isArray = false;
    let type = '', jsType = '', graphQLType = '';
    const modifiers = {};

    const arrayTypeChecker = /^\[([a-zA-Z0-9,{}]*)\](.*)$/;
    const singleTypeChecker = /^([a-zA-Z0-9,{}]*}?)(.*)$/;
    const enumTypeChecker = /^Enum{(.*)}(.*)$/;
    const arrayEnumTypeChecker = /^\[Enum{(.*)}\](.*)$/;

    // Separate type and modifiers
    const arrayEnumMatchData = tokens[1].match(arrayEnumTypeChecker);
    const enumMatchData = tokens[1].match(enumTypeChecker);
    if (tokens[1].match(arrayEnumTypeChecker)) {
      jsType = 'String';
      type = graphQLType = modelName +
        singularNestingContext.map((n) => capitalize(n)).join('') +
        (isArray ? capitalize(singular(name)) : capitalize(name));
      modifiers['enum'] = arrayEnumMatchData[1].split(',').map(quote);
      isArray = true;
      modifier = tokens[1].match(arrayEnumTypeChecker)[2];
    } else if (tokens[1].match(enumTypeChecker)) {
      jsType = 'String';
      type = graphQLType = modelName +
        singularNestingContext.map((n) => capitalize(n)).join('') +
        (isArray ? capitalize(singular(name)) : capitalize(name));
      modifiers['enum'] = enumMatchData[1].split(',').map(quote);
      modifier = tokens[1].match(enumTypeChecker)[2];
    } else if (tokens[1].match(arrayTypeChecker)) {
      isArray = true;
      modifier = tokens[1].match(arrayTypeChecker)[2];
      type = capitalize(tokens[1].match(arrayTypeChecker)[1]);
    } else if (tokens[1].match(singleTypeChecker)) {
      modifier = tokens[1].match(singleTypeChecker)[2];
      type = capitalize(tokens[1].match(singleTypeChecker)[1]);
    } else {
      throw `Unexpected type '${tokens[1]}'.`;
    }

    // Handle uploader types
    if (type.match(/.+Uploader$/)) {
      modifiers.uploader = type;
      jsType = graphQLType = type = 'File';
    }

    // String regexp match modifier
    if ((type === 'String') && modifier.match(/\/(.*)\/(.*)/)) {
      const matchData = modifier.match(/(\/.*\/[\w]*)(.*)/);
      const regex = matchData[1];
      modifier = matchData[2];
      modifiers.match = regex;
    }

    // min & max <= >= modifiers
    const minChecker = />=([\da-fx\.]*)/;
    const maxChecker = /<=([\da-fx\.]*)/;
    if (modifier.match(minChecker)) {
      modifiers[type === 'String' ? 'minlength' : 'min']
        = modifier.match(minChecker)[1];
      modifier = modifier.replace(minChecker, '');
    }
    if (modifier.match(maxChecker)) {
      modifiers[type === 'String' ? 'maxlength' : 'max']
        = modifier.match(maxChecker)[1];
      modifier = modifier.replace(maxChecker, '');
    }

    // Required
    if (modifier.match(/!/)) {
      modifiers.required = true;
    }

    // Index
    if (modifier.match(/\^/)) {
      modifiers.index = true;
    }

    // Unique
    if (modifier.match(/\$/)) {
      modifiers.unique = true;
    }

    // Handle sparse index
    if (modifiers.unique && !modifiers.required) {
      modifiers.sparse = true;
    }

    // Normalize js types and graphQL types
    if (primitiveTypes(ormSetting).includes(type)) {
      jsType = type;
      if (!graphQLType) {
        graphQLType = ormSetting.graphQLTypeFromORMType(type);
      }
    } else if (primitiveGraphQLTypes.includes(type)) {
      graphQLType = type;
      if (!jsType) {
        jsType = ormSetting.ormTypeFromGraphQLType(type);
      }
    } else {
      if (!jsType) jsType = type;
      if (!graphQLType) graphQLType = type;
    }

    // Primitive and reference types

    const primitive = primitiveTypes(ormSetting).includes(jsType);

    let foreignKey, foreignKeyIsArray;
    let assocModel, selfKey, destKey;

    // tokens[2] is default value for primitive types and
    // foreign key for reference types
    if (tokens[2]) {
      // Set default value for primitive types
      if (primitive) {
        // Primitive type
        if (tokens[2].match(/`(.*)`/)) {
          modifiers['default'] = tokens[2].match(/`(.*)`/)[1];
        } else {
          if (jsType === 'String') {
            modifiers['default'] = quote(tokens[2]);
          }
          if (jsType === 'Number') {
            modifiers['default'] = parseFloat(tokens[2]);
          }
          if (jsType === 'Boolean') {
            modifiers['default'] = JSON.parse(tokens[2]);
          }
          if (jsType === 'Date') {
            if (tokens[2].match(/^[0-9]+$/)) {
              modifiers['default'] = new Date(parseFloat(tokens[2]));
            } else {
              modifiers['default'] = new Date(tokens[2]);
            }
          }
        }
      } else {
        // Reference type with foreign key
        if (tokens[2].match(/^\[(.*)\]$/)) {
          foreignKeyIsArray = true;
          foreignKey = tokens[2].match(/^\[(.*)\]$/)[1];
        } else {
          if (tokens[2].match(/^[A-Z]/)) {
            // assoc model
            const assocInfo = tokens[2].split('.');
            assocModel = assocInfo[0];
            destKey = assocInfo[1] || uncapitalize(jsType);
            selfKey = assocInfo[2] || uncapitalize(modelName);
          } else {
            foreignKey = tokens[2];
          }
        }
      }
    }

    root(fields, nestingContext).push({
      name,
      type,
      jsType,
      graphQLType,
      isArray,
      primitive,
      modifiers,
      foreignKey,
      foreignKeyIsArray,
      assocModel,
      selfKey,
      destKey
    });

  });

  return removeUndefined({
    modelName,
    collectionName,
    varName,
    pluralVarName,
    fields
  });
};
